﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using TreeSharp;
using Styx.WoWInternals.WoWObjects;

namespace FelMaster.Classes.Priest
{
    public class DiscTest :HealerTalentBase
    {
        public override string KeySpell
        {
            get { return "Penance"; }
        }

        public override TreeSharp.Composite SingleRotation
        {
            get { return PVERotation; }
        }

        public override TreeSharp.Composite PVPRotation
        {
            get { return SingleRotation; }
        }

        private int GraceStacks { get { return SC.TargetCountBuff("Grace"); } }
        private bool CanShield { get { return !SC.TargetHasDebuff("Weakened Soul"); } }

        public override TreeSharp.Composite PVERotation
        {
            get {
                return new PrioritySelector(
                    new TreeSharp.Action(a => { SC.Log("In disco tree!"); return RunStatus.Failure; }),

                    //emergency heals on me
                    Healer.HealMe("Desperate Prayer", a => Me.HealthPercent < 40, "Desperate Prayer"),
                    //SC.HealMe("Flash Heal", a => Me.HealthPercent < 40, "flash heal on me, emergency"),

                    //don't break penance
                    new Decorator(x => SC.PlayerIsChanneling, new TreeSharp.Action()),

                    //emergency heals on most injured tank
                    Healer.FindTank(a => true, x => !x.Dead && !x.IsGhost && x.HealthPercent < 25, (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Power Word: Shield", a => CanShield && GraceStacks < 3, "Shield before penance (emergency)"),
                        SC.CastSpell("Penance", a => GraceStacks < 3, "Penance (emergency)"),
                        SC.CastSpell("Flash Heal", a => true, "Flash heal (emergency)")
                    ),

                    //I'm fine and tanks are not dying => ensure nobody is REALLY low life
                    Healer.FindRaidMember(a => true, x => !x.Dead && !x.IsGhost && x.HealthPercent < 10, (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Power Word: Shield", a => CanShield && GraceStacks < 3, "Shield before penance (emergency)"),
                        SC.CastSpell("Penance", a => GraceStacks < 3, "Penance (emergency)"),
                        SC.CastSpell("Flash Heal", a => true, "Flash heal (emergency)")
                    ),

                    //PoM on cooldown
                    Healer.FindTank(a => SC.SpellCooldown("Prayer Of Mending") < 0.5, x => !x.Dead && !x.IsGhost, (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Prayer of Mending", a => true, "PoM on tank")
                    ),

                    //party healing
                    Healer.FindParty(a => true, 10, 80, 30f, 4,
                        SC.CastSpell("Prayer Of Healing", a => true, "Prayer of healing")
                    ),

                    //single target healing
                    Healer.FindRaidMember(a => true, x => !x.Dead && !x.IsGhost && x.HealthPercent < 95, (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Penance", a => GraceStacks < 3, "Penance"),
                        SC.CastSpell("Greater Heal", a => Me.CurrentTarget.HealthPercent < 70, "Greater heal"),
                        SC.CastSpell("Heal", a => true, "heal")
                    ),

                    //cast shields while moving
                    Healer.FindRaidMember(a => Me.IsMoving && SC.CanCast("Power Word: Shield"), x => !x.Dead && !x.IsGhost && x.HealthPercent < 90 && !x.HasAura("Weakened Soul"), (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Power Word: Shield", a => true, "Shield while moving")
                    ),

                    //cast renew while moving
                    Healer.FindRaidMember(a => Me.IsMoving, x => !x.Dead && !x.IsGhost && x.HealthPercent < 90 && !x.ToPlayer().HasAura("Renew"), (a, b) => (int)(a.CurrentHealth - b.CurrentHealth),
                        SC.CastSpell("Renew", a => true, "Renew while moving")
                    )
                );
            }
        }

        public override TreeSharp.Composite Buffs
        {
            get
            {
                return new PrioritySelector(
                    SC.CastBuff("Inner Fire", a => true, "Inner Fire"),
                    SC.CastRaidBuff("Power Word: Fortitude", a => true, "Fortitude"),
                    SC.CastRaidBuff("Shadow Protection", a => true, "Shadow Protection")
                );
            }
        }

        public override string Name
        {
            get { return "Disc priest healing test"; }
        }
    }
}
